import pytest
from pathlib import Path
from braulio.git import _extract_commit_texts, Commit


@pytest.fixture(scope='session')
def fake_git_log_output():
    return Path('tests/data/commits.txt').read_text()


@pytest.fixture(scope='session')
def commit_text_list(fake_git_log_output):
    return _extract_commit_texts(fake_git_log_output)


@pytest.fixture(scope='session')
def commit_list(commit_text_list):
    return [Commit(text) for text in commit_text_list]


@pytest.fixture(scope='session')
def commit_text_registry(commit_text_list):
    return {c[7:14]: c for c in commit_text_list}


@pytest.fixture(scope='session')
def commit_registry(commit_list):
    return {commit.hash[:7]: commit for commit in commit_list}



import re


version_str_pattern = re.compile(
    '^(?:(?!0)\d+|0)[.](?:(?!0)\d+|0)[.](?:(?!0)\d+|0)$')


def validate_version_str(string):
    return bool(re.match(version_str_pattern, string))


class Version:
    def __init__(self, string=None, major=0, minor=0, patch=0):

        if string:
            self.string = string
            major, minor, patch = string.split('.')

        self.major = int(major)
        self.minor = int(minor)
        self.patch = int(patch)

        self._update_version_string()

    def bump(self, bump_type):
        if bump_type == 'major':
            self.major += 1
            self.minor, self.patch = 0, 0
        elif bump_type == 'minor':
            self.minor += 1
            self.patch = 0
        else:
            self.patch += 1

        self._update_version_string()

    def _update_version_string(self):
        self.string = f'{self.major}.{self.minor}.{self.patch}'

    def _is_comparable(self, value):
        if not isinstance(value, Version):
            raise TypeError(
                f"Comparison not supported"
                f" between instances of '{type(self)}'"
                f" and '{type(value)}'"
            )

    def __eq__(self, v):
        self._is_comparable(v)

        return (
            self.major == v.major and
            self.minor == v.minor and
            self.patch == v.patch
        )

    def __lt__(self, v):
        self._is_comparable(v)

        if self.major < v.major:
            return True

        if self.major > v.major:
            return False

        if self.minor < v.minor:
            return True

        if self.minor > v.minor:
            return False

        return self.patch < v.patch

    def __le__(self, v):
        self._is_comparable(v)

        return self < v or self == v

    def __str__(self):
        return self.string

    def __repr__(self):
        return f"Version('{self.string}')"


def get_next_version(bump_version_to, last_version=None):

    if bump_version_to not in {'major', 'minor', 'patch'}:
        new_version = Version(bump_version_to)

        if last_version and new_version <= last_version:
            return None

        return new_version

    new_version = Version()

    if last_version:
        new_version = Version(
            major=last_version.major,
            minor=last_version.minor,
            patch=last_version.patch,
        )

    new_version.bump(bump_version_to)

    return new_version



import click
from braulio.version import validate_version_str
from braulio.release import release as _release
from braulio.changelog import get_file_path, create_file


@click.group()
def cli():
    pass


@cli.command()
def init():
    click.echo('Initializing changelog')

    file_path = get_file_path()

    if not file_path:
        msg = 'No changelog file was found. Do you want to create a new one?'

        if click.confirm(msg):
            create_file()
    else:
        click.echo(f'{file_path.name} file found')


def bump_callback(ctx, param, value):
    if not value or validate_version_str(value):
        return value
    click.echo(f'{value} is not a valid version number')
    ctx.abort()


@cli.command()
@click.option('--major', 'bump_type', flag_value='major',
              help='Major version bump')
@click.option('--minor', 'bump_type', flag_value='minor',
              help='Minor version bump')
@click.option('--patch', 'bump_type', flag_value='patch',
              help='Patch version bump')
@click.option('--bump', 'bump_version_to', callback=bump_callback,
              is_eager=True)
@click.option('--commit/--no-commit', 'commit_flag', default=True)
@click.option('--tag/--no-tag', 'tag_flag', default=True)
@click.option('-y', 'y_flag', is_flag=True, default=False)
def release(bump_version_to, bump_type, commit_flag, tag_flag, y_flag):

    _release(
        bump_version_to=bump_version_to or bump_type,
        add_commit_flag=commit_flag,
        add_tag_flag=tag_flag,
        y_flag=y_flag,
    )



import re
from subprocess import run, PIPE
from braulio.version import Version

hash_pattern = re.compile('(?<=commit )\w{40}$', re.M)
task_pattern = re.compile('!\w+:\w*')


def _run_command(command):
    captured = run(command, stdout=PIPE)
    return captured.stdout.decode()


def _run_git_tag_command():
    command = [
        'git',
        'tag',
        '--sort=creatordate',
        '--format=%(creatordate:short)%09%(refname:strip=2)',
    ]
    return _run_command(command)


def _run_git_log_command(unreleased=False):
    command = ['git', 'log']

    if unreleased:
        tag_list = get_tags()

        if tag_list:
            last_tag = tag_list[0].name
            command.append(f'{last_tag}..HEAD')

    return _run_command(command)


def add_tag(tag_name):
    name = f'{tag_name}'
    return _run_command(['git', 'tag', '-a', name, '-m', '""'])


def add_commit(message):
    _run_command(['git', 'add', 'HISTORY.rst'])
    _run_command(['git', 'commit', '-m', f'"{message}"'])


def get_tags():

    tags_text = _run_git_tag_command()

    if tags_text == '':
        return []

    if tags_text[-1] == '\n':
        tags_text = tags_text[:-1]

    tag_text_list = tags_text.split('\n')
    tag_list = [Tag(text) for text in tag_text_list]

    return list(reversed(tag_list))


class Tag:
    def __init__(self, text):
        self.date = text[:10]
        self.name = re.search('v\d+[.]\d+[.]\d+', text).group(0)
        self.number = self.name[1:]

        self.version = Version(self.name[1:])

    def __str__(self):
        return f'Tag({self.version})'

    def __repr__(self):
        return f'Tag({self.version})'


class Commit:

    def __init__(self, text):
        # Remove EOL characters from the end of the text
        self.text = text[:-2] if text[-2:] == '\n\n' else text[:-1]

        text_lines = self.text.split('\n')

        # Commit hash
        self.hash = text_lines[0][7:]

        # Commit message
        msg_lines = [line[4:] for line in text_lines[4:]]
        self.message = '\n'.join(msg_lines)

        # Commit message header
        self.header = msg_lines[0]
        self.subject = self.header

        # Commit message body
        self.body = None
        self.scope = None
        self.action = None

        if len(msg_lines) > 2 and msg_lines[1] == '':
            self.body = '\n'.join(msg_lines[2:])

            match = re.search(task_pattern, msg_lines[-1])

            if match:
                task = match.group(0)
                action, scope = task.split(':')
                scope = None if scope == '' else scope

                self.action, self.scope = action[1:], scope

    def __repr__(self):
        return f'Commit(\'{self.header}\')'


patter = re.compile('^commit (?:.+?(?=commit \w{40})|.+$)', flags=re.M | re.S)


def _extract_commit_texts(git_log_text):
    return re.findall(patter, git_log_text)


def get_commits(unreleased=False):
    git_log_text = _run_git_log_command(unreleased=unreleased)
    commit_text_lst = _extract_commit_texts(git_log_text)

    return [Commit(commit_text) for commit_text in commit_text_lst]



import click
from datetime import date
from pathlib import Path


file_names = ('HISTORY.rst', 'CHANGELOG.rst',)
DEFAULT_FILE_NAME = file_names[0]


def get_file_path():

    for file_name in file_names:
        path = Path.cwd() / file_name

        if path.is_file():
            return path

    return None


def _make_title(title, level=1):
    underlines = ['=', '-', '~']
    underline = underlines[level - 1] * len(title)
    return f'{title}\n{underline}\n\n'


def create_file():
    path = (Path.cwd() / DEFAULT_FILE_NAME)
    path.touch()

    path.write_text(
        _make_title('History')
    )

    click.echo(f'{DEFAULT_FILE_NAME} created succesfully.')


def _make_sublist(commits):
    markup = ''

    for commit in commits:
        markup += f'  - {commit.subject}\n'

    return markup


def _make_list(scope_dict):
    markup = ''

    for commit in scope_dict['scopeless']:
        markup += f'* {commit.subject}\n'

    del scope_dict['scopeless']

    for scope_name, commit_lst in scope_dict.items():
        markup += f'* {scope_name}'

        if len(commit_lst) > 1:
            markup += '\n\n' + _make_sublist(commit_lst)
        else:
            markup += f' - {commit_lst[0].subject}\n'

    return markup + '\n'


def _make_release_markup(version, grouped_commits):
    today = str(date.today())
    title = f'{version.string} ({today})'
    markup = _make_title(title, level=2)

    if 'fix' in grouped_commits:
        markup += _make_title('Bug Fixes', level=3)
        markup += _make_list(grouped_commits['fix'])

    if 'feat' in grouped_commits:
        markup += _make_title('Features', level=3)
        markup += _make_list(grouped_commits['feat'])

    return markup


def update_changelog(version, grouped_commits):
    path = get_file_path()

    if not path:
        message = (
            'Unable to find a changelog file\n'
            'Run "$ brau init" to create one'
        )
        click.echo(message)
        return

    markup = _make_release_markup(version, grouped_commits)
    lines = []

    with path.open() as f:
        for line in f:
            lines.append(line)

    top = ''.join(lines[:3])
    bottom = ''.join(lines[3:])

    path.write_text(top + markup + bottom)



import click
from braulio.git import get_tags, get_commits, add_commit, add_tag
from braulio.changelog import update_changelog
from braulio.version import get_next_version


def _organize_commits(commit_list):
    by_action = {}
    has_breaking_changes = False

    for commit in commit_list:
        if commit.action:

            has_breaking_changes = (
                has_breaking_changes or 'BREAKING CHANGE' in commit.message
            )

            action = commit.action
            scope = commit.scope

            if action not in by_action:
                by_action[action] = {'scopeless': []}

            if not scope:
                by_action[action]['scopeless'].append(commit)
                continue
            else:
                if scope not in by_action[action]:
                    by_action[action][commit.scope] = []

                by_action[action][scope].append(commit)

    bump_type = 'major' if has_breaking_changes else \
                'minor' if 'feat' in by_action else \
                'patch'

    return {
        'bump_version_to': bump_type,
        'by_action': by_action,
    }


def release(
    bump_version_to=None, add_commit_flag=True, add_tag_flag=True, y_flag=False
):

    commit_list = get_commits(unreleased=True)

    if not commit_list:
        click.echo('Nothing to release')
        return False

    tag_list = get_tags()
    commits = _organize_commits(commit_list)
    bump_version_to = bump_version_to or commits['bump_version_to']
    last_version = tag_list[0].version if tag_list else None

    new_version = get_next_version(bump_version_to, last_version)

    if y_flag or click.confirm('Continue?'):

        update_changelog(
            version=new_version,
            grouped_commits=commits['by_action'],
        )

        if add_commit_flag:
            add_commit(f'Release version {new_version.string}')

        if add_tag_flag:
            add_tag('v' + new_version.string)



import pytest
from unittest.mock import patch
from datetime import date
from pathlib import Path
from click.testing import CliRunner
from braulio.git import Commit, Version
from braulio.release import _organize_commits
from braulio.changelog import _make_title, _make_sublist, _make_list, \
    _make_release_markup, update_changelog


parametrize = pytest.mark.parametrize


@parametrize(
    'level, expected',
    [
        (1, 'Title\n=====\n\n'),
        (2, 'Title\n-----\n\n'),
        (3, 'Title\n~~~~~\n\n'),
    ],
)
def test_make_title(level, expected):
    output = _make_title('Title', level)
    assert output == expected


def test_make_sublist(commit_text_registry):
    reg = commit_text_registry
    commits = [
        Commit(reg['80a9e0e']),
        Commit(reg['0a10908']),
        Commit(reg['eaedb93']),
    ]

    markup = _make_sublist(commits)

    assert markup == (
        f'  - {commits[0].subject}\n'
        f'  - {commits[1].subject}\n'
        f'  - {commits[2].subject}\n'
    )


def test_make_list(commit_text_registry):
    reg = commit_text_registry

    commits = [
        Commit(reg['80a9e0e']),
        Commit(reg['0a10908']),
        Commit(reg['eaedb93']),
    ]

    scope_dict = {
        'scopeless': [
            commits[0],
            commits[1],
        ],
        'scope1': [
            commits[0],
            commits[1],
        ],
        'scope2': [commits[1]],
        'scope3': [
            commits[0],
            commits[1],
        ],
    }

    markup = _make_list(scope_dict)

    assert markup == (
        f'* {commits[0].subject}\n'
        f'* {commits[1].subject}\n'
        '* scope1\n\n'
        f'  - {commits[0].subject}\n'
        f'  - {commits[1].subject}\n'
        f'* scope2 - {commits[1].subject}\n'
        '* scope3\n\n'
        f'  - {commits[0].subject}\n'
        f'  - {commits[1].subject}\n\n'
    )


class Test_make_release_markup:

    def test_release_with_fixes_and_features(self, commit_list):
        version = Version(major=10, minor=3, patch=0)
        commits = _organize_commits(commit_list)
        markup = _make_release_markup(
            version,
            grouped_commits=commits['by_action'],
        )

        assert markup == (
            f'10.3.0 ({str(date.today())})\n'
            '-------------------\n\n'
            'Bug Fixes\n'
            '~~~~~~~~~\n\n'
            '* thing - Fix a thing\n\n'
            'Features\n'
            '~~~~~~~~\n\n'
            '* Add a thing\n'
            '* music\n\n'
            '  - Add more music please\n'
            '  - Add cool musics :D\n'
            '* cli - Add a cli tool\n\n'
        )

    def test_release_with_fixes(self, commit_registry):
        reg = commit_registry
        commit_list = [reg['4d17c1a']]
        commits = _organize_commits(commit_list)
        version = Version(major=10, minor=3, patch=0)

        markup = _make_release_markup(
            version,
            grouped_commits=commits['by_action'],
        )

        assert markup == (
            f'10.3.0 ({str(date.today())})\n'
            '-------------------\n\n'
            'Bug Fixes\n'
            '~~~~~~~~~\n\n'
            '* thing - Fix a thing\n\n'
        )

    def test_release_with_features(self, commit_registry):
        reg = commit_registry
        version = Version(major=10, minor=3, patch=0)
        commit_list = [reg['ccaa185'], reg['bc0bcab'], reg['a6b655f']]
        commits = _organize_commits(commit_list)
        markup = _make_release_markup(
            version,
            grouped_commits=commits['by_action'],
        )

        assert markup == (
            f'10.3.0 ({str(date.today())})\n'
            '-------------------\n\n'
            'Features\n'
            '~~~~~~~~\n\n'
            '* Add a thing\n'
            '* music - Add more music please\n'
            '* cli - Add a cli tool\n\n'
        )


class TestUpdateChangelog:

    @patch('braulio.changelog.click.echo')
    def test_missing_changelog_file(self, mocked_echo):
        runner = CliRunner()

        with runner.isolated_filesystem():
            version = Version()

            update_changelog(version, {})

            mocked_echo.assert_called_with(
                'Unable to find a changelog file\n'
                'Run "$ brau init" to create one'
            )

    @patch('braulio.changelog._make_release_markup')
    def test_write_to_changelog_file(
        self, mock_make_release_markup
    ):
        mock_make_release_markup.return_value = 'New Content'
        runner = CliRunner()

        with runner.isolated_filesystem():
            with open('CHANGELOG.rst', 'w') as f:
                f.write(
                    f'{_make_title("History")}'
                    'Content'
                )

            version = Version()
            organized_commits = {}

            update_changelog(version, organized_commits)

            text = Path('CHANGELOG.rst').read_text()

            assert text == (
                'History\n'
                '=======\n\n'
                'New Content'
                'Content'
            )

            mock_make_release_markup.assert_called_with(
                version, organized_commits
            )



import pytest
from unittest.mock import patch, Mock
from pathlib import Path
from click.testing import CliRunner
from braulio.cli import cli
from braulio.git import Tag, Version


parametrize = pytest.mark.parametrize


class TestInit:

    def test_when_changelog_file_is_not_found(self):
        runner = CliRunner()

        with runner.isolated_filesystem():
            result = runner.invoke(cli, ['init'])

            assert 'Do you want to create a new one?' in result.output
            assert '[y/N]:' in result.output

    def test_create_changelog_file(self):

        runner = CliRunner()

        with runner.isolated_filesystem():
            path = Path.cwd() / 'HISTORY.rst'
            assert not path.exists()

            runner.invoke(cli, ['init'], input='y')

            assert path.exists()

    @pytest.mark.parametrize(
        'file_name',
        ['HISTORY.rst', 'CHANGELOG.rst'],
    )
    def test_known_changelog_files(self, file_name):
        runner = CliRunner()

        with runner.isolated_filesystem():
            with open(file_name, 'w') as f:
                f.write('')

            result = runner.invoke(cli, ['init'])

            assert f'{file_name} file found' in result.output


class TestRelease:

    @patch('braulio.release.get_commits', return_value=[])
    def test_no_commits(self, mocked_get_commits):
        runner = CliRunner()

        result = runner.invoke(cli, ['release'])

        assert 'Nothing to release' in result.output
        mocked_get_commits.assert_called_with(unreleased=True)

    @patch('braulio.release.get_commits')
    @patch('braulio.release.add_tag')
    @patch('braulio.release.add_commit')
    @patch('braulio.release.update_changelog')
    def test_pass_no_to_confirmation_prompt(
        self,
        mocked_update_changelog,
        mocked_add_commit,
        mocked_add_tag,
        mocked_get_commits
    ):
        runner = CliRunner()

        result = runner.invoke(cli, ['release'])

        assert result.exit_code == 0
        assert 'Continue? [y/N]' in result.output

        mocked_update_changelog.assert_not_called()
        mocked_add_commit.assert_not_called()
        mocked_add_tag.assert_not_called()

    @patch('braulio.release.get_commits')
    @patch('braulio.release.add_tag')
    @patch('braulio.release.add_commit')
    @patch('braulio.release.update_changelog')
    def test_pass_yes_to_confirmation_prompt(
        self,
        mocked_update_changelog,
        mocked_add_commit,
        mocked_add_tag,
        mocked_get_commits,
    ):
        runner = CliRunner()

        result = runner.invoke(cli, ['release'], input='y')
        assert result.exit_code == 0
        assert 'Continue? [y/N]' in result.output

        mocked_update_changelog.assert_called()
        mocked_add_commit.assert_called()
        mocked_add_tag.assert_called()

    @parametrize(
        'option, tags, expected',
        [
            ('--major', [Tag('2015-10-15 v0.10.1')], '1.0.0',),
            ('--minor', [Tag('2015-10-15 v0.10.1')], '0.11.0',),
            ('--patch', [Tag('2015-10-15 v0.10.1')], '0.10.2',),
            ('--major', [], '1.0.0',),
            ('--minor', [], '0.1.0',),
            ('--patch', [], '0.0.1',),
            ('--bump=4.0.0', [Tag('2015-10-15 v2.0.1')], '4.0.0',),
            ('--bump=3.0.0', [], '3.0.0',),
        ],
    )
    @patch('braulio.release.add_tag')
    @patch('braulio.release.add_commit')
    @patch('braulio.release.update_changelog')
    @patch('braulio.release.get_tags')
    @patch('braulio.release.get_commits')
    def test_manual_version_bump(
        self,
        mocked_get_commits,
        mocked_get_tags,
        mocked_update_changelog,
        mocked_add_commit,
        mocked_add_tag,
        option,
        tags,
        expected,
        commit_list,
    ):
        mocked_get_commits.return_value = commit_list
        mocked_get_tags.return_value = tags

        runner = CliRunner()
        result = runner.invoke(cli, ['release', option], input='y')

        assert result.exit_code == 0

        mocked_update_changelog.assert_called()

        # Check what version was passed to update_changelog function
        passed_args = mocked_update_changelog.call_args[1]
        assert passed_args['version'] == Version(expected)

        mocked_add_commit.assert_called_with(f'Release version {expected}')
        mocked_add_tag.assert_called_with(f'v{expected}')

    @parametrize(
        'hash_lst, tags, expected',
        [
            (['4d17c1a', '80a9e0e'], [Tag('2016-10-03 v0.7.6')], '0.7.7'),
            (['4d17c1a', 'ccaa185'], [Tag('2016-10-03 v0.7.6')], '0.8.0'),
            (['8c8dcb7', 'ccaa185'], [Tag('2016-10-03 v0.7.6')], '1.0.0'),
            (['4d17c1a', '80a9e0e'], None, '0.0.1'),
            (['4d17c1a', 'ccaa185'], None, '0.1.0'),
            (['8c8dcb7', 'ccaa185'], None, '1.0.0'),
        ],
    )
    @patch('braulio.release.add_tag')
    @patch('braulio.release.add_commit')
    @patch('braulio.release.update_changelog')
    @patch('braulio.release.get_tags')
    @patch('braulio.release.get_commits')
    def test_determine_next_version_from_commit_messages(
        self,
        mocked_get_commits,
        mocked_get_tags,
        mocked_update_changelog,
        mocked_add_commit,
        mocked_add_tag,
        hash_lst,
        tags,
        expected,
        commit_registry,
    ):
        mocked_get_tags.return_value = tags
        mocked_get_commits.return_value = [
            commit_registry[short_hash] for short_hash in hash_lst
        ]

        runner = CliRunner()

        result = runner.invoke(cli, ['release', '-y'])

        assert result.exit_code == 0, result.exception

        mocked_update_changelog.assert_called()

        # Check what version was passed to update_changelog function
        passed_args = mocked_update_changelog.call_args[1]
        assert passed_args['version'] == Version(expected)

        mocked_add_commit.assert_called_with(f'Release version {expected}')
        mocked_add_tag.assert_called_with(f'v{expected}')

    @patch('braulio.release._organize_commits')
    @patch('braulio.release.add_tag')
    @patch('braulio.release.add_commit')
    @patch('braulio.release.update_changelog')
    @patch('braulio.release.get_tags')
    @patch('braulio.release.get_commits')
    @patch('braulio.release.get_next_version')
    def test_update_changelog(
        self,
        mocked_get_next_version,
        mocked_get_commits,
        mocked_get_tags,
        mocked_update_changelog,
        mocked_add_commit,
        mocked_add_tag,
        mocked_organize_commits,
    ):
        runner = CliRunner()
        result = runner.invoke(cli, ['release', '-y'])

        assert result.exit_code == 0

        mocked_get_commits.assert_called_with(unreleased=True)

        mocked_organize_commits.assert_called_with(
            mocked_get_commits()
        )

        mocked_get_next_version.assert_called_with(
            mocked_organize_commits()['bump_version_to'],
            mocked_get_tags()[0].version
        )

        mocked_update_changelog.assert_called_with(
            version=mocked_get_next_version(),
            grouped_commits=mocked_organize_commits()['by_action']
        )

    @patch('braulio.release.add_tag')
    @patch('braulio.release.add_commit')
    @patch('braulio.release.update_changelog')
    @patch('braulio.release.get_tags', return_value=[])
    @patch('braulio.release.get_commits')
    def test_no_commit_flag(
        self,
        mocked_get_commits,
        mocked_get_tags,
        mocked_update_changelog,
        mocked_add_commit,
        mocked_add_tag,
        commit_list,
    ):
        mocked_get_commits.return_value = commit_list

        runner = CliRunner()
        result = runner.invoke(cli, ['release', '--no-commit', '-y'])

        assert result.exit_code == 0

        mocked_add_commit.assert_not_called()
        mocked_add_tag.assert_called()

    @patch('braulio.release.add_tag')
    @patch('braulio.release.add_commit')
    @patch('braulio.release.update_changelog')
    @patch('braulio.release.get_tags', return_value=[])
    @patch('braulio.release.get_commits')
    def test_no_tag_flag(
        self,
        mocked_get_commits,
        mocked_get_tags,
        mocked_update_changelog,
        mocked_add_commit,
        mocked_add_tag,
        commit_list,
    ):
        mocked_get_commits.return_value = commit_list

        runner = CliRunner()
        result = runner.invoke(cli, ['release', '--no-tag', '-y'])

        assert result.exit_code == 0

        mocked_add_commit.assert_called()
        mocked_add_tag.assert_not_called()



import pytest
from braulio.release import _organize_commits


class TestOrganizedCommits:

    @pytest.mark.parametrize(
        'hash_list, expected',
        [
            (['eaedb93', '8c8dcb7'], 'major'),
            (['ccaa185', '264af1b'], 'minor'),
            (['4d17c1a', '80a9e0e'], 'patch'),
        ],
    )
    def test_bump_version_to(self, commit_registry, hash_list, expected):
        commit_list = [
            commit_registry[short_hash] for short_hash in hash_list
        ]
        commits = _organize_commits(commit_list)
        assert commits['bump_version_to'] == expected

    def test_commit_grouping_by_action_and_scope(self, commit_list):
        commits = _organize_commits(commit_list)

        grouped_commits = commits['by_action']

        assert len(grouped_commits['fix']['thing']) == 1
        assert len(grouped_commits['feat']['scopeless']) == 1
        assert len(grouped_commits['feat']['cli']) == 1
        assert len(grouped_commits['feat']['music']) == 2
        assert len(grouped_commits['refactor']['music']) == 1
        assert len(grouped_commits['refactor']['lorem']) == 1



import pytest
from braulio.version import validate_version_str, Version, get_next_version


parametrize = pytest.mark.parametrize


@parametrize(
    'string, expected',
    [
        ('1', False),
        ('1.0', False),
        ('a.b.c', False),
        ('1.1.a', False),
        ('12345', False),
        ('abcd', False),
        ('01.0.1', False),
        ('1.01.1', False),
        ('1.0.01', False),
        ('1.1.1.1', False),
        ('1.1.1', True),
        ('1.10.1', True),
        ('1.0.0', True),
        ('0.1.0', True),
        ('0.0.1', True),
    ]
)
def test_validate_version_str(string, expected):
    assert validate_version_str(string) is expected


class TestVersion:

    def test_constructor_without_arguments(self):
        version = Version()

        assert version.major == 0
        assert version.minor == 0
        assert version.patch == 0
        assert version.string == '0.0.0'

    def test_constructor_with_major_minor_patch_arguments(self):
        version = Version(major=1, minor=2, patch=3)

        assert version.major == 1
        assert version.minor == 2
        assert version.patch == 3
        assert version.string == '1.2.3'

    @parametrize(
        'bump_type, major, minor, patch, string',
        [
            ('major', 2, 0, 0, '2.0.0'),
            ('minor', 1, 3, 0, '1.3.0'),
            ('patch', 1, 2, 4, '1.2.4'),
        ]
    )
    def test_bump_method(self, bump_type, major, minor, patch, string):
        version = Version('1.2.3')
        version.bump(bump_type=bump_type)

        assert version.major == major
        assert version.minor == minor
        assert version.patch == patch
        assert version.string == string

    @parametrize(
        'left, right, expected',
        [
            ('4.0.0', '4.0.0', False),
            ('2.0.1', '2.0.0', False),
            ('1.4.0', '1.3.0', False),
            ('3.0.0', '2.23.8', False),
            ('2.0.0', '2.0.1', True),
            ('1.3.0', '1.4.0', True),
            ('2.23.8', '3.0.0', True),
        ]
    )
    def test_less_than_operator(self, left, right, expected):
        assert (Version(left) < Version(right)) is expected

    @parametrize(
        'left, right, expected',
        [
            ('4.0.0', '4.0.0', False),
            ('2.0.1', '2.0.0', True),
            ('1.4.0', '1.3.0', True),
            ('3.0.0', '2.23.8', True),
            ('2.0.0', '2.0.1', False),
            ('1.3.0', '1.4.0', False),
            ('2.23.8', '3.0.0', False),
        ]
    )
    def test_greater_than_operator(self, left, right, expected):
        assert (Version(left) > Version(right)) is expected

    @parametrize(
        'left, right, expected',
        [
            ('2.24.1', '2.24.1', True),
            ('7.4.0', '2.23.8', False),
            ('1.3.0', '1.4.0', False),
            ('2.23.8', '3.0.0', False),
        ]
    )
    def test_equal_to_operator(self, left, right, expected):
        assert (Version(left) == Version(right)) is expected

    @parametrize(
        'left, right, expected',
        [
            ('2.24.1', '2.24.1', False),
            ('7.4.0', '2.23.8', True),
            ('1.3.0', '1.4.0', True),
            ('2.23.8', '3.0.0', True),
        ]
    )
    def test_no_equal_to_operator(self, left, right, expected):
        assert (Version(left) != Version(right)) is expected

    @parametrize(
        'left, right, expected',
        [
            ('2.24.1', '2.24.1', True),
            ('7.4.0', '2.23.8', False),
            ('1.3.0', '1.4.0', True),
            ('2.23.8', '3.0.0', True),
        ]
    )
    def test_less_or_equal_to_operator(self, left, right, expected):
        assert (Version(left) <= Version(right)) is expected

    @parametrize(
        'left, right',
        [
            (Version(), 1),
            (Version(), '2.23.8'),
            (Version(), None),
            (Version(), True),
            (1, Version()),
            ('2.23.8', Version()),
            (None, Version()),
            (True, Version()),
        ]
    )
    def test_comparators_with_invalid_values(self, left, right):
        error_message = 'not supported'

        with pytest.raises(TypeError, match=error_message):
            left > right

        with pytest.raises(TypeError, match=error_message):
            left < right

        with pytest.raises(TypeError, match=error_message):
            left <= right

        with pytest.raises(TypeError, match=error_message):
            left >= right

        with pytest.raises(TypeError, match=error_message):
            left == right

        with pytest.raises(TypeError, match=error_message):
            left != right


@parametrize(
    'bump_version_to, last_version, expected',
    [
        ('0.0.0', Version('0.0.0'), None),
        ('0.0.0', Version('0.0.1'), None),
        ('0.0.2', Version('0.0.1'), Version('0.0.2')),
        ('0.0.0', None, Version('0.0.0')),
        ('7.0.0', None, Version('7.0.0')),
        ('0.8.0', None, Version('0.8.0')),
        ('0.0.12', None, Version('0.0.12')),
        ('major', Version('4.4.7'), Version('5.0.0')),
        ('minor', Version('4.4.7'), Version('4.5.0')),
        ('patch', Version('4.4.7'), Version('4.4.8')),
        ('major', None, Version('1.0.0')),
        ('minor', None, Version('0.1.0')),
        ('patch', None, Version('0.0.1')),
    ]
)
def test_get_next_version(bump_version_to, last_version, expected):
    new_version = get_next_version(
        bump_version_to=bump_version_to,
        last_version=last_version,
    )

    assert new_version == expected



import pytest
from subprocess import PIPE
from unittest.mock import patch, Mock, call
from braulio.git import _run_git_tag_command, _run_git_log_command, \
    get_tags, get_commits, _run_command, add_tag, add_commit, \
    Commit, Tag


parametrize = pytest.mark.parametrize


@patch('braulio.git.run')
def test_run_command(mocked_run):
    mocked_capture = Mock()
    mocked_capture.stdout.decode.return_value = 'output value'
    mocked_run.return_value = mocked_capture

    output = _run_command(['command', 'subcomand'])

    assert output == 'output value'
    mocked_run.assert_called_with(['command', 'subcomand'], stdout=PIPE)


@patch('braulio.git._run_command')
def test_run_git_tag(mocked_run_command):
    mocked_run_command.return_value = 'Stdout text'

    result = _run_git_tag_command()

    mocked_run_command.assert_called_with(
        [
            'git',
            'tag',
            '--sort=creatordate',
            '--format=%(creatordate:short)%09%(refname:strip=2)',
        ],
    )
    assert result == 'Stdout text'


@patch('braulio.git._run_command')
def test_add_tag(mocked_run_command):
    add_tag('tagname')

    mocked_run_command.assert_called_with(
        ['git', 'tag', '-a', 'tagname', '-m', '""'],
    )


@patch('braulio.git._run_command')
def test_add_commit(mocked_run_command):
    add_commit('A message')

    assert mocked_run_command.call_count == 2
    assert mocked_run_command.call_args_list[0] == call(
        ['git', 'add', 'HISTORY.rst'],
    )
    assert mocked_run_command.call_args_list[1] == call(
        ['git', 'commit', '-m', '"A message"'],
    )


@parametrize(
    'unreleased, command, tags',
    [
        (False, ['git', 'log'], [Tag('2015-10-15 v0.0.1')]),
        (False, ['git', 'log'], []),
        (True, ['git', 'log', 'v0.0.1..HEAD'], [Tag('2015-10-15 v0.0.1')]),
        (True, ['git', 'log'], []),
    ]
)
@patch('braulio.git.get_tags')
@patch('braulio.git._run_command')
def test_run_git_log_command(
    mocked_run_command,
    mocked_get_tags,
    unreleased,
    command,
    tags
):
    mocked_run_command.return_value = 'Stdout text'
    mocked_get_tags.return_value = tags

    result = _run_git_log_command(unreleased=unreleased)

    mocked_run_command.assert_called_with(command)
    assert result == 'Stdout text'


class Test_get_commits:

    @patch('braulio.git._run_git_log_command')
    def test_get_all_commits(self, mock_run_git_log, fake_git_log_output):
        mock_run_git_log.return_value = fake_git_log_output

        commits = get_commits()

        mock_run_git_log.assert_called_with(unreleased=False)
        assert len(commits) == 12

        first = commits[0]
        assert first.header == 'Fix lorem ipsum dolor sit amet'

        last = commits[-1]
        assert last.header == 'Add additional information (#26)'

    @patch('braulio.git._run_git_log_command')
    def test_get_unreleased_commits(
        self,
        mock_run_git_log,
        fake_git_log_output
    ):

        mock_run_git_log.return_value = fake_git_log_output

        commits = get_commits(unreleased=True)

        mock_run_git_log.assert_called_with(unreleased=True)
        assert len(commits) == 12

        first = commits[0]
        assert first.header == 'Fix lorem ipsum dolor sit amet'

        last = commits[-1]
        assert last.header == 'Add additional information (#26)'


@parametrize(
    'text, date, name',
    [
        ('2015-10-15      v10.0.1', '2015-10-15', 'v10.0.1'),
        ('2016-02-26      v0.10.13', '2016-02-26', 'v0.10.13'),
        ('2016-07-06      v0.0.5', '2016-07-06', 'v0.0.5'),
        ('2016-10-03      v0.70.6', '2016-10-03', 'v0.70.6'),
    ]
)
def test_tag_class(text, date, name):
    tag = Tag(text)

    assert tag.name == name
    assert tag.date == date


class Test_get_tags:

    @patch('braulio.git._run_git_tag_command')
    def test_without_tags(self, mock_run_git_tag):
        mock_run_git_tag.return_value = ''
        tag_list = get_tags()
        assert tag_list == []

    @patch('braulio.git._run_git_tag_command')
    def test_with_tags(self, mock_run_git_tag):
        mock_run_git_tag.return_value = (
            '2015-10-15      v0.0.1\n'
            '2015-11-18      v0.0.2\n'
            '2016-02-26      v0.0.3\n'
            '2016-03-22      v0.0.4\n'
            '2016-07-06      v0.0.5\n'
            '2016-10-03      v0.0.6\n'
        )

        tag_list = get_tags()

        mock_run_git_tag.assert_called()

        for _tag in tag_list:
            assert type(_tag) == Tag

        assert len(tag_list) == 6
        assert tag_list[0].name == 'v0.0.6'
        assert tag_list[5].name == 'v0.0.1'


class TestCommit:

    @parametrize(
        'short_hash, expected',
        [
            ('eaedb93', 'eaedb9320c7aad581daa05f1510b64393c082dbb'),
            ('80a9e0e', '80a9e0edf6208b3c1c9bc28bfb1a5f04b1f59e1d'),
        ],
        ids=[
            'eaedb9320c7aad581daa05f1510b64393c082dbb',
            '80a9e0edf6208b3c1c9bc28bfb1a5f04b1f59e1d',
        ],
    )
    def test_hash(self, commit_text_registry, short_hash, expected):
        text = commit_text_registry[short_hash]
        assert Commit(text).hash == expected

    @parametrize(
        'short_hash, expected',
        [
            ('eaedb93', 'Fix lorem ipsum dolor sit amet'),
            ('80a9e0e', 'Stop being lazy'),
            ('2f9be68', 'Add more lorem ipsum text'),
            ('0a10908', 'Add additional information (#26)'),
        ],
        ids=[
            'with body',
            'without line break',
            'without body',
            'with multiple paragraphs',
        ],
    )
    def test_header(self, commit_text_registry, short_hash, expected):
        text = commit_text_registry[short_hash]
        assert Commit(text).header == expected

    @parametrize(
        'short_hash',
        ['80a9e0e', '2f9be68'],
        ids=['no line break before header', 'without body'],
    )
    def test_message_with_invalid_body(self, commit_text_registry, short_hash):
        text = commit_text_registry[short_hash]
        assert Commit(text).body is None

    @parametrize(
        'short_hash, expected',
        [
            (
                'eaedb93',
                (
                    'Consectetur adipiscing elit, sed do\n'
                    'eiusmod tempor incididunt ut labore et\n'
                    'dolore magna aliqua.'
                )
            ),
            (
                '82f0d12',
                (
                    'What? Make it yourself.\n'
                    '\n'
                    'Sudo make me a sandwich.'
                )
            ),
            (
                '0a10908',
                (
                    'But I must explain to you how all this mistaken\n'
                    'idea of denouncing pleasure and praising pain was\n'
                    'born.\n'
                    '\n'
                    'Who avoids a pain that produces no resultant pleasure?.'
                )
            ),
        ],
        ids=[
            'Single paragraph',
            'multiple single line paragraphs',
            'multiple paragraphs',
        ],
    )
    def test_messages_with_valid_body(
        self,
        commit_text_registry,
        short_hash,
        expected
    ):
        text = commit_text_registry[short_hash]
        assert Commit(text).body == expected

    @parametrize(
        'last_line, action, scope',
        [
            ('Random text', None, None),
            ('!feat:cli', 'feat', 'cli'),
            ('!feat:cli and some more text', 'feat', 'cli'),
            ('Closed #2 and fixed #3 !feat:cli', 'feat', 'cli'),
            ('!feat:', 'feat', None),
            ('!feat: and some more text', 'feat', None),
            ('Closed #2 and fixed #3 !feat:', 'feat', None),
        ],
    )
    def test_action_and_scope_lookup(self, last_line, action, scope):

        text = (
            'commit eaedb9320c7aad581daa05f1510b64393c082dbb\n'
            'Author: René Descartes <rene.decartes@xyz.test>\n'
            'Date:   Thu Apr 19 14:11:05 2018 +0100\n'
            '\n'
            '    Fix lorem ipsum dolor sit amet\n'
            '\n'
            '    Consectetur adipiscing elit, sed do\n'
            '    eiusmod tempor incididunt ut labore et\n'
            '    dolore magna aliqua.\n'
            '\n'
            f'    {last_line}\n\n'
        )

        commit = Commit(text)

        assert commit.scope == scope
        assert commit.action == action
